##############################################################################
#
# RROC-casestudies.r: part of the RROCcurves library
#
# RROCcurves library
#  JOSE HERNANDEZ-ORALLO, UNIVERSITAT POLITECNICA DE VALENCIA, SPAIN
#  jorallo@dsic.upv.es
#
# See "INFO.TXT" for copyright information, history, related papers, etc.
#
##############################################################################



#####################
#### DIRECTORIES ####

# Directories where the other .R files must be found and the PDF files will be output. 
# Customise this for your computer and OS.
WORKDIR <- "A:/__FAENA__/_RESEARCH/_SUBMISSIONS/2012/Regression ROC Curves/script/RROCcurves" # casa
# WORKDIR <- "E:/__FAENA__/_RESEARCH/_SUBMISSIONS/2012/Regression ROC Curves/script" # poli
setwd(WORKDIR)

source("RROC-functions.R")



################################
#
# CODE ORGANISATION
#
# This file includes example with real datasets
#
# First, we have some functions. These functions call regression methods such as LR, svm, regression trees, etc., and need packages
# Second, we have the part to load the dataset and split it into work dataset and deployment dataset
# Third, there is code where we work with the work dataset for several repetitions (work dataset) and plot many curves
# Fourth, there is code to analyse the models and plot hem for the deployment dataset
#
###############################











##################################
#
#   SOME FUNCTIONS
#   These are just part of the experiment, not really "exportable/public" functions of the RROCcurves library
#
##################################


# There is a function "dist" in the dist package which implements all this.

# Euclidean distance
distEuc <- function(v1, v2) {
  v <- 0
  for(i in 1:length(v1)) {
    if (is.numeric(v1[i])) {
      v <- v + (v1[i] - v2[i])^2
    } else if (v1[i] != v2[i]) {
      v <- v + 1
    }
  }

  sqrt(sum(v))
}

# Manhattan distance
distMan <- function(v1, v2) {
  v <- 0
  for(i in 1:length(v1)) {
    if (is.numeric(v1[i])) {
      v <- v + abs(v1[i] - v2[i])
    } else if (v1[i] != v2[i]) {
      v <- v + 1
    }
  }

  sum(v)  
}



# K-nearest neighbour algorithm
knnpredict <- function(train, etest, k, WEIGHTED, INCLUDE_EQUALS, mydist, apply_log=FALSE) {
# Assumes the class is the last attribute
# WEIGHTED == 0 means no weight, 1 means the inverse of the distance is used, 2 is squared, etc.
# INCLUDE_EQUALS == TRUE considers the examples that are exactly the same as the example that wants to be predicted (set it to TRUE if you do not know what it means)

  nattr <- length(train[1,])
  nexamp <- length(train[,1])

  distvect <- rep(0, nexamp)  # distance vector to all the examples
  for(i in 1:nexamp) {
    etrain <- train[i, ]  
      distvect[i] <- mydist(etrain[1:(nattr-1)],etest[1:(nattr-1)])
  }
  indices <- sort(distvect, index.return= TRUE)

  k <- min(nexamp, k) # in case there are fewer examples than k

  distvect <- rep(0, k)  # distance vector of the k best
  classvect <- rep(0, k)  # class vector of the k best
  weights <- rep(0, k)
  num <- 0
  den <- 0
  for (i in 1:k) {
    distvect[i] <- indices$x[i]
    classvect[i] <- train[indices$ix[i], nattr]  # takes the last attribute of the involved example

    if (WEIGHTED > 0) {
      if (distvect[i] == 0) # this may lead to divisions by 0
        weights[i] <- Inf
      else
        weights[i] <- 1/(distvect[i]^WEIGHTED) 
    }


  }

  m0 <- classvect[1]  # I keep the first one just in case I have no examples at the end)

  if (INCLUDE_EQUALS) {
    # check for 0 distances
    if (WEIGHTED > 0) {  # treat 0 distances (only if they make some weight infinite)
      classvect <- classvect[which(weights != Inf)] 
      weights <- weights[which(weights != Inf)] 
    }
  }  else {
    classvect <- classvect[which(distvect != 0)] 
    weights <- weights[which(distvect != 0)] 
  }

  if (apply_log) {
    classvect <- log(classvect)
  } 

  if (WEIGHTED > 0) {
    m <- weighted.mean(classvect, weights)
  } else {
    m <- mean(classvect)
  }

  if (apply_log) {
    m <- exp(m)
  } 


  if ((k == 1) || (length(classvect) <=1)) {
    m <- m0
  }
             
  m
}




# Builds several models given a training dataset and predictd them on the test set.
BuildModels <- function(train.df, test.df, apply_log= FALSE, build_KNN= TRUE) {
# apply_log: whether the output variable is transformed with a log
# By default it creates LR, SVM, tree and a mean model. Optionally (build_KNN==TRUE) also creates a kNN model

    nattr <- length(train.df[1,])
    len <- length(test.df[,1])
  
    # Constructing models with train.df and getting predictions with test.df

  if (apply_log) {
    lmmodel <- lm(log(class) ~ ., data= train.df)      #    a linear model
    svmmodel <- ksvm(log(class) ~ ., data= train.df)   #    a svm
    trmodel <- rpart(log(class) ~ ., data= train.df)   #    a tree

    pred_tm <- exp(rep(mean(log(train.df[,nattr])), len))
    pred_lr <- exp(predict(lmmodel, test.df))   # , se.fit=TRUE)$fit
    pred_svm <- exp(predict(svmmodel, test.df)) #, se.fit=TRUE)$fit
      pred_svm <- pred_svm[1:length(pred_svm)]  # flattens the vector
    pred_tr <- exp(predict(trmodel, test.df, se.fit=TRUE, type="vector"))

  } else {
    lmmodel <- lm(class ~ ., data= train.df)      #    a linear model
    svmmodel <- ksvm(class ~ ., data= train.df)   #    a svm
    trmodel <- rpart(class ~ ., data= train.df)   #    a tree

    pred_tm <- rep(mean(train.df[,nattr]), len)
    pred_lr <- predict(lmmodel, test.df) # , se.fit=TRUE)$fit
    pred_svm <- predict(svmmodel, test.df) #, se.fit=TRUE)$fit
      pred_svm <- pred_svm[1:length(pred_svm)]  # flattens the vector
    pred_tr <- predict(trmodel, test.df, se.fit=TRUE, type="vector")
  }

  my_list <- list(pred_tm, pred_lr, pred_svm, pred_tr)

  if (build_KNN) {
    pred_knn <- NULL
    for (i in 1:len) {
      pred_knn[i] <- knnpredict(train, test.df[i, ], k=10, WEIGHTED= 0, TRUE, mydist, apply_log)
    }

    my_list <- c(my_list, list(pred_knn))
  }

  my_list
  
}  




# Plots several models passed as a list compared to the actual values
PlotModels <- function(model_list, actual, pdfdatasetname, name_vector = NULL, col_vector = NULL, lty_vector = NULL, REC_ERROR_FUNCTION= 1) {
# model_list: a list of models, each one represented as an array of predictions

  lenlist <- length(model_list)

  # Ensures that the lists of colors, ltyles and names match the list of models

  lencol <- length(col_vector)
  lenlty <- length(lty_vector)
  lenname <- length(name_vector)

  if (lencol == 0) {
    col_vector <- rep("grey", lenlist)
  } else if (lencol < lenlist) {
    col_vector <- rep(col_vector, trunc((lenlist-1) / lencol) + 1)
  }

  if (lenlty == 0) {
    lty_vector <- rep(1, lenlist)
  } else if (lenlty < lenlist) {
    lty_vector <- rep(lty_vector, trunc((lenlist-1) / lenlty) + 1)
  }

  if (lenname == 0) {
    name_vector <- rep("", lenlist)
  } else  if (lenname < lenlist) {
    name_vector <- rep(name_vector, trunc((lenlist-1) / lenname) + 1)
  }


  # Model representation as a density plot of the predicted and actual values

  cat("\n\nDENSITY PLOTS\n")

  if (PDFOPEN) {
    pdf(paste(pdfdatasetname,"DENS", ".pdf", sep=""), height= PDFheight, width= PDFwidth)
  }

  plot(density(actual), main="")

  for (i in 1:lenlist) {
    pred  <- model_list[[i]]  
    lines(density(pred),col=col_vector[i], lty=lty_vector[i])
  }

#  smartlegend(x="right",y="top", legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)
  legend(x="topright",cex= 0.8, legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)

  if (PDFCLOSE) {
    dev.off()
  }


  # Model representation as a density plot of the residuals

  cat("\n\nRESIDUAL DENSITY PLOTS\n")

  if (PDFOPEN) {
    pdf(paste(pdfdatasetname,"RESDENS", ".pdf", sep=""), height= PDFheight, width= PDFwidth)
  }

  density_list <- NULL
  ymax <- 0
  xmin <- Inf
  xmax <- -Inf
  for (i in 1:lenlist) {
    pred  <- model_list[[i]]  
    values <- pred - actual
    if (i==1) {
      my_dens <- density(values)
      my_bw <- my_dens$bw
    } else {
      my_dens <- density(values, bw=my_bw)
    }
    ymax <- max(ymax, my_dens$y)
    xmin <- min(xmin, my_dens$x)
    xmax <- max(xmax, my_dens$x)

    density_list <- c(density_list, list(my_dens))
  }

  for (i in 1:lenlist) {
    if (i==1) {
      plot(density_list[[i]],col=col_vector[i], lty=lty_vector[i], ylim=c(0,ymax), xlim=c(xmin,xmax), main="")
    } else {
      lines(density_list[[i]],col=col_vector[i], lty=lty_vector[i])
    }
  }


#  smartlegend(x="left",y="top",legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)
  legend(x="topleft",cex= 0.8, legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)

  if (PDFCLOSE) {
    dev.off()
  }



  # Model representation as a scatterplot

  cat("\n\nSCATTER PLOTS\n")

  if (PDFOPEN) {
    pdf(paste(pdfdatasetname,"SCAT", ".pdf", sep=""), height= PDFheight, width= PDFwidth)
  }

  xlim0 <- min(actual)
  xlim1 <- max(actual)
  ylim0 <- xlim0
  ylim1 <- xlim1
  plot(0, 0, lty="blank", col="white", xlim=c(xlim0,xlim1), ylim=c(ylim0,ylim1), xlab="actual", ylab="predicted")  

  for (i in 1:lenlist) {
    pred  <- model_list[[i]]  
    points(actual,pred,col=col_vector[i], pch=lty_vector[i])  # lty=lty_vector[i])
  }

#  smartlegend(x="left",y="top",legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)
  legend(x="topleft",cex= 0.8, legend=name_vector, col=col_vector, pch=lty_vector, text.col=col_vector)  #lty=lty_vector, text.col=col_vector)
  
  if (PDFCLOSE) {
    dev.off()
  }



  # Model representation with REC curves

  cat("\n\nREC PLOTS\n")

  if (REC_ERROR_FUNCTION == 1) {
    errorfun <- abserror
    recname <- "REC-ABS"
  } else {
    errorfun <- squerror
    recname <- "REC-SQU"
  }
  
  if (PDFOPEN) {
    pdf(paste(pdfdatasetname, recname,  ".pdf", sep=""), height= PDFheight, width= PDFwidth)
  }

  # Calculates the maximum error to know the width of the x-axis
  predall <- NULL
  actualrep <- NULL
  for (i in 1:lenlist) {
    predall <- c(predall,model_list[[i]])
    actualrep <- c(actualrep, actual)
  }
  xlim <- max(errorfun(predall,actualrep)) * 0.5   # Half of the maximum value
  
  REC_space(xlim, REC_ERROR_FUNCTION)

  for (i in 1:lenlist) {
    pred  <- model_list[[i]]  
    cat("\n\nModel: ", name_vector[i], "\n")
    RECCURVE <- REC_curve(pred, actual, errorfun, plot_curve=TRUE, col_vector[i], DRAW_CIRCLES= FALSE, my_lty=lty_vector[i])
  }

#  smartlegend(x="right",y="bottom",legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)
  legend(x="bottomright",cex= 0.8, legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)

  if (PDFCLOSE) {
    dev.off()
  }



  # Model representation with RROC curves

  cat("\n\nRROC PLOTS\n")

  EXAMPLES_DOMINANCE <- FALSE # If this is true, some sample intercepts are shown and the two worst models removed (for the dataset used in the paper)

  if (PDFOPEN) {
    if (EXAMPLES_DOMINANCE) {
      pdf(paste(pdfdatasetname,"RROCdom", ".pdf", sep=""), height= PDFheight, width= PDFwidth)
    } else {
      pdf(paste(pdfdatasetname,"RROC", ".pdf", sep=""), height= PDFheight, width= PDFwidth)
    }
  }

  NORMALISE_AXES <- TRUE
  N <- length(actual)

  factor <- 0.2
  if (EXAMPLES_DOMINANCE) {  # We remove the last two models from the plot because they are worse
    lenlist <- 3
    factor <- 0.15
  }

  PLOT_DIAGONAL <- !EXAMPLES_DOMINANCE
#  if (NEWPLOT) {
    xlim0 <- data_range * factor
    ylim0 <- data_range * factor 
    if (NORMALISE_AXES) {  # Because the function below makes the division we need to multiply here
      xlim0 <- xlim0 * N
      ylim0 <- ylim0 * N
    }
    RROC_space(NORMALISE_AXES, N, xlim0, ylim0, PLOT_DIAGONAL)
#  }

  PLOT_POINT <- !EXAMPLES_DOMINANCE

  PLOT_RECTANGLE <- FALSE
  PLOT_CURVE <- TRUE
  

  curve_list <- list()
  for (i in 1:lenlist) {
    pred  <- model_list[[i]]  
    cat("\n\nModel: ", name_vector[i], "\n")
    RROCCURVE <- RROC_curve(pred, actual, NULL, PLOT_POINT, PLOT_RECTANGLE, PLOT_CURVE, col_vector[i], col_vector[i], NULL, normalise_axes=NORMALISE_AXES, N, xlim0, ylim0, DRAW_CIRCLES= FALSE, my_lty=lty_vector[i])
    curve_list <- c(curve_list,list(RROCCURVE))
  }

#  smartlegend(x="right",y="bottom",legend=name_vector, col=col_vector, lty=lty_vector, text.col=col_vector)
  legend(x="bottomright",cex= 0.8, legend=name_vector[1:lenlist], col=col_vector[1:lenlist], lty=lty_vector[1:lenlist], text.col=col_vector[1:lenlist])

  if (EXAMPLES_DOMINANCE) {
    # This code is just to include some intercepts for the figures on the paper
    alpha <- 0.05
    RROC_intercept_many(curve_list, alpha, plot_isometric=TRUE, mycol="lightgrey", normalise_axes=NORMALISE_AXES, N)

    alpha <- 0.25
    RROC_intercept_many(curve_list, alpha, plot_isometric=TRUE, mycol="lightgrey", normalise_axes=NORMALISE_AXES, N)

    alpha <- 0.75
    RROC_intercept_many(curve_list, alpha, plot_isometric=TRUE, mycol="lightgrey", normalise_axes=NORMALISE_AXES, N)
  }

  if (PDFCLOSE) {
    dev.off()
  }

  curve_list
}






















################################################################
#
#   PACKAGES NEEDED
#
################################################################







  # install.packages(...)

  require("tree") || install.packages("tree")   # For tree modelling
  library(tree) # package for CART modelling

  require("rpart") || install.packages("rpart")   # For tree modelling
  library(rpart) # package for CART modelling

#  require("neuralnet") || install.packages("neuralnet")   # For neural net modelling
#  library(neuralnet) 

#  require("nnet") || install.packages("nnet")   # For neural net modelling
#  library(nnet) 

  require("kernlab") || install.packages("kernlab")   # For ksvm
  library(kernlab) 

#  require("gplots") || install.packages("gplots")   # For smartlegend
#  library(gplots)

  #  library(MASS) # for lm.ridge










#################################################################
#
#   CHOOSING THE DATASET
#
#   Uncomment one or the other depending on what you want to use
#
#################################################################


  pred_given <- NULL  # pred_given is used occasionally for datasets where there is a reference, expert or human model.


#    require("np") || install.packages("np")   # For some datasets
#    library(np)
#    data(oecdpanel)   
#    class <- oecdpanel[,1]
#    dat <- cbind(oecdpanel[,2:7], class)

#    library(MASS)    
#    data(Boston)     # Boston buroughs, models get different results and much better than the mean model
#    dat <- Boston

#    data(birthwt)    # Birth weight
#    dat <- birthwt



#    library(foreign)  # read.arff
#    dat <- read.arff("strike.arff")  # models get different results and much better than the mean model


#    dat <- read.arff("concrete.arff")  
#    http://archive.ics.uci.edu/ml/datasets/Concrete+Compressive+Strength   


#   dat <- read.csv("californiahousing.csv", header = TRUE, sep = " ")  # MOLT GRAN I NO CREUEN!
#   class <- dat[,1]
#    dat <- cbind(dat[,2:9], class)


#   library(Ecdat) 
#   data(Housing)                     # the models do not cross
#   dat <- Housing
#   class <- dat[,1]
#   dat <- cbind(dat[,2:12], class)

#    library(Ecdat) 
#    pdfdatasetname <- "Clothing"
#    data(Clothing)   # Clothing shops in the Netherlands http://127.0.0.1:24884/library/Ecdat/html/Clothing.html
#    dat <- Clothing
#    # Noting that dat[1]/dat[2] == dat[12] we have to take column 1 as class remove column 2
#     class <- dat[,1]
#    dat <- cbind(dat[,3:length(dat[1,])], class)   # We remove the two first columns
#    pred_given <- NULL

#    pdfdatasetname <- "kuiper"
#    dat <- read.csv("kuiper.csv")  # About car prices
#    class <- dat[,1]  / 1000     # We divide by 1000 to ease visualisation        # The SVM and kNN cross nicely and LR gets behind (this does not happen now once the numerisation is well done). But kNN is too slow
#    class <- log(class)          # A reasonable transformation
#    dat <- cbind(dat[,2:length(dat[1,])], class)
#    # http://www.amstat.org/publications/jse/v16n3/datasets.kuiper.html
#    # Some nominal values have many labels: [4] has 32, [5] has 47. This is problematic (we may need some preprocessing here)
#    # Remove these attributes
#    dat <- dat[,c(1:3,6:length(dat[1,]))]


#     dat <- read.csv("cars.csv", na.strings = "?", stringsAsFactors=FALSE)
#     # 8 RetailPrice
#     # 9 DealerCost
#     dat <- data.matrix(dat)  # Converts dat to a numeric matrix (or ensures that it is numeric)
#  # ANOTHER OPTION IS TO REMOVE LINEAR REGRESSION SINCE IT DOES NOT TAKE NULL VALUES.
#  
#  
#  #   dat <- na.omit(dat)      # Removes NA values if there are
#     x <- dat[,13]
#     x[is.na(x)] = median(x,TRUE)
#     dat[,13] <- x
#     x <- dat[,14]
#     x[is.na(x)] = median(x,TRUE)
#     dat[,14] <- x
#     x <- dat[,15]
#     x[is.na(x)] = median(x,TRUE)
#     dat[,15] <- x
#     x <- dat[,16]
#     x[is.na(x)] = median(x,TRUE)
#     dat[,16] <- x
#     x <- dat[,17]
#     x[is.na(x)] = median(x,TRUE)
#     dat[,17] <- x
#     x <- dat[,18]
#     x[is.na(x)] = median(x,TRUE)
#     dat[,18] <- x
#   
#     class <- dat[,8]   # RetailPrice
#  
#     pred_given <- dat[,9]
#  
#     # lm(class ~ pred_given)
#     # (Intercept)   pred_given  
#     #  -255.8          1.1  
#     lo <- loess(class ~ pred_given)
#     pred_given <- predict(lo, pred_given)
#  #   pred_given <- pred_given * 1.1 - 255.8   
#  
#  #   perc <- mean(pred_given /dat[,8])
#  #   pred_given <- pred_given / perc # LLEVAR-LI UN PERCENTATGE
#  
#     dat <- cbind(dat[,c(1:7,10:length(dat[1,]))], class)  # Both 8 and 9 are removed
#  # http://www.amstat.org/publications/jse/jse_data_archive.htm


# This one is used in the article and provided with the library
    pdfdatasetname <- "swCSC"
    dat <- read.csv("swCSC.csv")  # About software effort    
    # Those labels with less than 10 values for an attribute are merged (affects attributes 1 and 2)
    dat <- dat[-102,]                    # Example 102 is an outlier
    class <- dat[,6]  # Actual Effort
    pred_given <- dat[,4]  # First.estimate
    APPLY_LOG <- FALSE
    if (APPLY_LOG == TRUE) {
      dat[,3] <- log(dat[,3])  # If we use a logarithmic transformation, we also need to transform the adjusted function points 
      pdfdatasetname <- "LOGswCSC"   # I put LOG to have LOGLOG to indicate that the input has also been transformed
    }
    dat <- cbind(dat[,c(1,2,3)], class)   # Only Client.code, Project.Type and Adjusted.function.points

    # 5 repetitions, seed: 100  BE
    # 10 repetitions, seed: 2222 MOLT BE









################################################################
#
#   OPTIONS (about plots and the experiment itself  
#
################################################################

# APPLY_LOG <- FALSE # That depends on the dataset. Should be enabled or disabled above

#### Initialises errorfun for REC curves (not used for RROC curves)
REC_ERROR_FUNCTION <- 1  # 1 MAE, 2 MSE

NORMALISE_AXES <- TRUE    # Divide x-axis and y-axis by n (OVER/n and UNDER/n)

BUILD_KNN <- FALSE
DIST <- 1
if (DIST == 1) {   # This choice affects the kNN (if used) and the SD methods (if used)
  mydist <- distEuc 
} else {
  mydist <- distMan
}

# NAME_VECTOR <- c(expression("m"[Mean]), expression("m"[LR]), expression("m"[SVM]), expression("m"[Tree]), expression("m"[kNN]), expression("m"[given]))
if (BUILD_KNN) {
  NAME_VECTOR <- c("Mean", "LR", "SVM", "Tree", "kNN", "Given")
} else {
  NAME_VECTOR <- c("Mean", "LR", "SVM", "Tree", "Given")
}
COL_VECTOR <- c("cyan", "red", "blue","green","violet","orange", "grey")
LTY_VECTOR <- c(1,2,4,5,6,3)

CROSS_VALIDATION <- TRUE
REPETITIONS <- 10

# Options for the RCOST curves
selname <- "Selection"
sellty <- 3
selcol <- "black"
sellwd <- 3


# PDF options
PDFOPEN <- TRUE           # If the plots are output on a PDF file
PDFCLOSE <- PDFOPEN        # Close the PDF file. This should match PDFOPEN, except when you want to draw several curves before closing.
PDFheight= 5.5             # 7 is the default, so 14 makes it double higher than wide, 5 makes letters bigger (in proportion) for just one plot
PDFwidth= 5                # (as above for width) 7 by default
# mymar <- c(5, 4, 4, 2) + 0.1  # default
# mymar <- c(5, 5, 4, 2) + 0.1  # you can use this to modify the margins of your PDF file
# par(mar=mymar) 












################################################################
#
#   Arranging and processing the dataset "dat" 
#
################################################################


############## SCALING, RANGES, ... #################

if (APPLY_LOG) {
   pdfdatasetname <- paste("LOG", pdfdatasetname, sep="")
}


# dat <- data.matrix(dat)  # Converts dat to a numeric matrix (or ensures that it is numeric)
                         # But it spoils the nominal attributes   

                        
dat <- na.omit(dat)      # Removes NA values if there are

datalen <- length(dat[,1])
nattr <- length(dat[1,])


cat("\n\n\nWorking with a dataset", pdfdatasetname, "of length:", datalen, "and number of attributes:", nattr, "\n\n")


for (i in 1:(nattr-1)) {
  if (is.numeric(dat[1,i])) {  #Does not affect nominal attributes if there are
    dat[,i] <- scale(dat[,i])  # Scale the input attributes (this is useful for some methods, especially kNN) 
  }
}

data_range <- max(dat[,nattr]) - min(dat[,nattr])
min_range <- min(dat[,nattr])
data_sd <- sd(dat[,nattr])



###### SPLIT BETWEEN WORK DATA AND DEPLOYMENT DATA #####

set.seed(2222) # We fix the seed for repeatability of results 

# SHUFFLE!!!
shufindx <- sample(1:datalen, datalen)  
dat <- dat[shufindx,]

DEPLOYPERC <- 0.33
cut <- trunc(datalen*(1-DEPLOYPERC))
datwork <- dat[1:cut,]
if (!is.null(pred_given))
  pred_given_work <- pred_given[1:cut]
dataworklen <- cut
datdeploy <- dat[(cut+1):datalen,]
if (!is.null(pred_given))
  pred_given_deploy <- pred_given[(cut+1):datalen]

















################################################################
#
#   REPETITIONS WITH THE WORK DATA : TRAINING AND PLOTTING 
#
################################################################


####### WE START THE REPETITIONS WITH THE WORK DATA #######


cat("\n\n\nAfter the split, we do", REPETITIONS, "repetitions\n\n")


pred_tmALL <- NULL
pred_lrALL <- NULL
pred_svmALL <- NULL
pred_trALL <- NULL
if (BUILD_KNN) 
  pred_knnALL <- NULL
pred_givenALL <- NULL
actualALL <- NULL

for(repe in 1:REPETITIONS) {

 print("Repetition:")
 print(repe)

  if (CROSS_VALIDATION) {
    foldlen <- dataworklen/REPETITIONS

    if (foldlen < 1) {
      print("CROSS-VALIDATION ERROR: folds too small")
      kkkkkk
    }

    itest0 <- round((repe-1)*foldlen+1)
    itest1 <- round((repe)*foldlen)
    test <- datwork[itest0:itest1, , drop=FALSE]   # DROP = FALSE is to prevent this being converted into a vector (instead of a matrix) when the size of the testset is 1.
    if (repe > 1) {
      train1 <- datwork[1:round((repe-1)*foldlen), ]
    }
    if (repe < REPETITIONS) {
      train2 <- datwork[round((repe)*foldlen+1):dataworklen, ]
    }
    if (repe == 1) {
      train <- train2
    } else if (repe == REPETITIONS) {
      train <- train1
    } else {
      train <- rbind(train1, train2)
    }
    trainlen <- length(train[,1])

    if (!is.null(pred_given)) {
      pred_given_test <- pred_given_work[itest0:itest1]
    }
  } else {
    # SHUFFLE!!!  (for each repetition)
    shufindx <- sample(1:dataworklen, dataworklen)  
    datwork <- datwork[shufindx,]

    trainlen <- round(dataworklen * TRAINPERC) 
    train <- datwork[1:trainlen,]
    test <- datwork[(trainlen+1):dataworklen, , drop=FALSE]   # DROP = FALSE is to prevent this being converted into a vector (instead of a matrix) when the size of the testset is 1.
    if (!is.null(pred_given)) {
      pred_given_test <- pred_given_work[(trainlen+1):dataworklen]
    }
  }
  testlen <- length(test[,nattr])

  train_mean <- mean(train[,nattr])
  train_sd <- sd(train[,nattr])

  actual <- test[,nattr] 
  test_mean <- mean(actual)


  # Constructing dataframes for the methods.
  train.df <- data.frame(x=train[,1:(nattr-1)], class= train[,nattr])
  test.df <- data.frame(x=test[,1:(nattr-1), drop=FALSE])  # This works (just in case the test has one element).
  if (nattr <= 2) {  # If there is only one input attribute we need to fix this as well
    colnames(test.df)[1] <- "x"
  }

  model_list <- BuildModels(train.df, test.df, APPLY_LOG, BUILD_KNN)
  if (!is.null(pred_given))
    model_list <- c(model_list, list(pred_given_test))


  PlotModels(model_list, actual, paste(pdfdatasetname, "REP", repe, sep=""), NAME_VECTOR, COL_VECTOR, LTY_VECTOR, REC_ERROR_FUNCTION) 

  pred_tmALL <- c(pred_tmALL,model_list[[1]])
  pred_lrALL <- c(pred_lrALL,model_list[[2]])
  pred_svmALL <- c(pred_svmALL,model_list[[3]])
  pred_trALL <- c(pred_trALL,model_list[[4]])
  if (BUILD_KNN) {
    pred_knnALL <- c(pred_knnALL,model_list[[5]])
  }
  pred_givenALL <- c(pred_givenALL,pred_given_test)
  actualALL <- c(actualALL,actual)

} # END OF MAIN EXPERIMENTAL LOOP ON THE WORK DATA







################################################################
#
#   PROCESSING ALL THE REPETITIONS TOGETHER: MORE TRAINING AND PLOTTING
#
################################################################


## With all the iterations we plot the RROC curves

  cat("\n\n\nITERATIONS FINISHED\nWith all of them we plot density/REC/RROC plots for the work data\n\n")


  model_list1 <- list(pred_tmALL, pred_lrALL, pred_svmALL, pred_trALL)

  if (BUILD_KNN) {
    model_list1 <- c(model_list1, list(pred_knnALL))
  }

  if (!is.null(pred_given))
    model_list1 <- c(model_list1, list(pred_givenALL))
  name_vector <- c(expression("m"[Mean]), expression("m"[LR]), expression("m"[SVM]), expression("m"[Tree]), expression("m"[kNN]), expression("m"[given]))
  curve_list1 <-   PlotModels(model_list1, actualALL, paste(pdfdatasetname, "ALL", sep=""), NAME_VECTOR, COL_VECTOR, LTY_VECTOR, REC_ERROR_FUNCTION) 


## and calculate dominance intervals

  dom_intervals1 <- RROC_dominance_intervals(curve_list1, RESOLUTION)
  print(dom_intervals1)


# Calculate the optimal shifts with the work dataset for the range of alphas 

  NUM_ALPHAS <- 201
  alpha_vect <- seq(0,1,length.out=NUM_ALPHAS)
  
  actual <- actualALL

  shift_vect_list <- NULL
  for(i in 1:length(model_list1)) {
    predicted <- model_list1[[i]]
    shift_vect <- NULL
    for (alpha in alpha_vect) {
      res <- optimal_shift_linlin(predicted, actual, alpha)
      shift_vect <- c(shift_vect, res)
    }
    shift_vect_list <- c(shift_vect_list, list(shift_vect))
  }


# Apply the optimal shifts (calculated with the work dataset) for the range of alphas on the work dataset

  if (PDFOPEN) {
    pdf(paste(pdfdatasetname,"ALL", "RCOST",  ".pdf", sep=""), height= PDFheight, width= PDFwidth)
  }

  normalise_axes <- TRUE
  N <- length(actual)

  for(i in 1:length(model_list1)) {
    predicted <- model_list1[[i]]
    shift_vect <- shift_vect_list[[i]]

    if (i==1) {
      refloss <- loss_linlin(predicted,actual,0.5)  # We use one arbitrary loss as a reference
      RCOST_space_linlin(refloss*1.5,normalise_axes, N)               # And make the space half higher than that (it is an approximation). Change the factor if the y-axis is too high or too low.     
    }

    RCOST_curve_linlin(predicted, actual, alpha_vect, shift_vect, plot_curve= TRUE, colour2=COL_VECTOR[i], DRAW_CIRCLES=FALSE, my_lty=LTY_VECTOR[i], my_lwd=1, normalise_axes, N) 
  }

# Now we highlight the regions chosen as dominance intervals

  for(j in 1:length(dom_intervals1)) {

    # We recover the region
    res <- dom_intervals1[[j]]
    alpha0 <- res[1]  # Interval start
    abline(v=alpha0, col="lightgrey", lty=3)
    alpha1 <- res[2]  # Interval end
    abline(v=alpha1, col="lightgrey", lty=3)

    # We recover the model
    i <- res[3]  # Choose the model
    predicted <- model_list1[[i]]
    shift_vect <- shift_vect_list[[i]]

    # We just restrict to the region
    range <- (alpha_vect >= alpha0) & (alpha_vect <= alpha1)
    alpha_vect_j <- alpha_vect[range]
    shift_vect_j <- shift_vect[range]
    
    RCOST_curve_linlin(predicted, actual, alpha_vect_j, shift_vect_j, plot_curve= TRUE, colour2=selcol, DRAW_CIRCLES=FALSE, my_lty=sellty, my_lwd=sellwd, normalise_axes, N) 
  }

  newcolours <- c(COL_VECTOR[1:length(model_list1)],selcol)
#  smartlegend(x="center",y="bottom",legend=c(NAME_VECTOR,selname), col=newcolours, lty=c(LTY_VECTOR,sellty), lwd=c(rep(1,length(model_list1)),sellwd), text.col=newcolours)
  legend(x="bottom",cex= 0.8, legend=c(NAME_VECTOR,selname), col=newcolours, lty=c(LTY_VECTOR,sellty), lwd=c(rep(1,length(model_list1)),sellwd), text.col=newcolours)


  if (PDFCLOSE) {
    dev.off()
  }










################################################################
#
#   WE NOW CHANGE TO THE DEPLOYMENT DATASET: MORE TRAINING AND PLOTTING
#
################################################################



############ NOW WE EVALUATE THINGS WITH THE DEPLOYMENT DATA

  cat("\n\n\nWe now plot density/REC/RROC plots for the deployment data\n\n")


## Now we train with all the working data and evaluate with the deployment data.

  # Constructing dataframes for the methods.
  trainall.df <- data.frame(x=datwork[,1:(nattr-1)], class= datwork[,nattr])
  deploy.df <- data.frame(x=datdeploy[,1:(nattr-1), drop=FALSE])  # This works (just in case the test has one element).
  if (nattr <= 2) {  # If there is only one input attribute we need to fix this as well
    colnames(deploy.df)[1] <- "x"
  }

  model_list2 <- BuildModels(trainall.df, deploy.df, APPLY_LOG, BUILD_KNN)
  actual <- datdeploy[,nattr]
  if (!is.null(pred_given)) {
    model_list2 <- c(model_list2, list(pred_given_deploy))
  }
  curve_list2 <- PlotModels(model_list2, actual, paste(pdfdatasetname, "DEPLOY", sep=""), NAME_VECTOR, COL_VECTOR, LTY_VECTOR, REC_ERROR_FUNCTION) 

## and calculate dominance intervals

  dom_intervals2 <- RROC_dominance_intervals(curve_list2, RESOLUTION)
  print(dom_intervals2)


  cat("\n\n\nWithout the selection\n\n")


# Apply the optimal shifts (calculated with the work dataset) for the range of alphas on the deployment dataset

  if (PDFOPEN) {
    pdf(paste(pdfdatasetname,"DEPLOY", "RCOST",  ".pdf", sep=""), height= PDFheight, width= PDFwidth)
  }


  normalise_axes <- TRUE
  N <- length(actual)

  for(i in 1:length(model_list2)) {
    predicted <- model_list2[[i]]
    shift_vect <- shift_vect_list[[i]]

    if (i==1) {
      refloss <- loss_linlin(predicted,actual,0.5)  # We use one arbitrary loss as a reference
      RCOST_space_linlin(refloss*1.5,normalise_axes, N)               # And make the space half higher than that (it is an approximation). Change the factor if the y-axis is too high or too low.     
    }

    RCOST_curve_linlin(predicted, actual, alpha_vect, shift_vect, plot_curve= TRUE, colour2=COL_VECTOR[i], DRAW_CIRCLES=FALSE, my_lty=LTY_VECTOR[i], my_lwd=1, normalise_axes, N) 
  }



# The same process but now we highlight the regions chosen as dominance intervals (the dominance intervals are those recognised on the work dataset)

  cat("\n\n\nSame process with the selection\n\n")

  for(j in 1:length(dom_intervals1)) {

    # We recover the region
    res <- dom_intervals1[[j]]
    alpha0 <- res[1]  # Interval start
    abline(v=alpha0, col="lightgrey", lty=3)
    alpha1 <- res[2]  # Interval end
    abline(v=alpha1, col="lightgrey", lty=3)

    # We recover the model
    i <- res[3]  # Choose the model
    predicted <- model_list2[[i]]
    shift_vect <- shift_vect_list[[i]]

    # We just restrict to the region
    range <- (alpha_vect >= alpha0) & (alpha_vect <= alpha1)
    alpha_vect_j <- alpha_vect[range]
    shift_vect_j <- shift_vect[range]
    
    RCOST_curve_linlin(predicted, actual, alpha_vect_j, shift_vect_j, plot_curve= TRUE, colour2=selcol, DRAW_CIRCLES=FALSE, my_lty=sellty, my_lwd=sellwd, normalise_axes, N) 
  }

  newcolours <- c(COL_VECTOR[1:length(model_list2)],selcol)
#  smartlegend(x="center",y="bottom",legend=c(NAME_VECTOR,selname), col=newcolours, lty=c(LTY_VECTOR,sellty), lwd=c(rep(1,length(model_list2)),sellwd), text.col=newcolours)
  legend(x="bottom",cex= 0.8, legend=c(NAME_VECTOR,selname), col=newcolours, lty=c(LTY_VECTOR,sellty), lwd=c(rep(1,length(model_list2)),sellwd), text.col=newcolours)


  if (PDFCLOSE) {
    dev.off()
  }
